import csv
import logging
import os
import tempfile
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from ego4d.cli.manifest import list_videos_in_manifest, VideoMetadata
from ego4d.cli.universities import BUCKET_TO_UNIV, UNIV_TO_BUCKET
from ego4d.internal.ffmpeg_utils import get_video_info, VideoInfo
from ego4d.internal.university_files import *
import collections
import logging
import math
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from threading import Lock

import boto3
import botocore
import pandas as pd
from botocore.exceptions import ClientError
from ego4d.internal.credential_s3 import S3Helper
from tqdm import tqdm

lock = Lock()

"""
-i: The S3 path where the metadata is stored
-a: check all of the latest files for each university
-mf: "./ego4d/internal/standard_metadata_v10"
-ed: "error_details"
-es: "error_summary"
"""


def _validate_video_components(
    s3,
    bucket_name,
    video_metadata_dict: Dict[str, VideoMetadata],
    video_components_dict: Dict[str, List[VideoComponentFile]],
    error_message: List[ErrorMessage],
) -> List[Tuple[str, str]]:
    """
    Args:
        video_components_dict: Dict[str, List[VideoComponentFile]]:
        mapping from university_video_id to a list of all VideoComponentFile
        objects that have equal values at university_video_id

        error_message: List[ErrorMessage]: a list to store ErrorMessage objects
        generated when validating video components.

    Returns:
        List[Tuple[str, str]] a list containing a tuple for each
        video component's university_video_id and key
    """
    print("validating video components")
    video_info_param = []
    for video_id, components in tqdm(video_components_dict.items()):
        if video_id not in video_metadata_dict:
            error_message.append(
                ErrorMessage(
                    video_id,
                    "video_not_found_in_video_metadata_error",
                    f"{video_id} in video_components_dict can't be found in video_metadata",
                )
            )
            continue
        elif video_metadata_dict[video_id].number_video_components != len(components):
            error_message.append(
                ErrorMessage(
                    video_id,
                    "video_component_length_inconsistent_error",
                    f"the video has {len(components)} components when it"
                    f" should have {video_metadata_dict[video_id].number_video_components}",
                )
            )
        components.sort(key=lambda x: x.component_index)
        # check component_index is incremental and starts at 0
        university_video_folder_path = video_metadata_dict[
            video_id
        ].university_video_folder_path
        for i in range(len(components)):
            if components[i].component_index != i:
                error_message.append(
                    ErrorMessage(
                        video_id,
                        "video_component_wrong_index_error",
                        f"the video component has index {components[i].component_index}"
                        f" when it should have {i}",
                    )
                )
            if not components[i].video_component_relative_path:
                error_message.append(
                    ErrorMessage(
                        video_id,
                        "empty_video_component_relative_path_error",
                        f"Found an empty relative path for the video_id {video_id}",
                    )
                )
            else:
                s3_path = f"{university_video_folder_path.rstrip('/')}/{components[i].video_component_relative_path}"
                bucket, key = split_s3_path(s3_path)
                if bucket != bucket_name:
                    error_message.append(
                        ErrorMessage(
                            video_id,
                            "bucket_name_inconsistent_error",
                            f"video has bucket_name {bucket_name}"
                            f" when it should have {bucket}",
                        )
                    )
                try:
                    s3.head_object(Bucket=bucket, Key=key)
                    video_info_param.append((video_id, key))
                except ClientError as e:
                    error_message.append(
                        ErrorMessage(
                            video_id,
                            "path_does_not_exist_error",
                            f"video s3://{bucket}/{key} doesn't exist in bucket",
                        )
                    )

    return video_info_param


def _get_videos(
    s3,
    bucket_name: str,
    video_info_param: List[Tuple[str, str]],
    error_message: List[ErrorMessage],
) -> Dict[str, List[VideoInfo]]:
    """
    Args:
        video_info_param: List[Tuple[str, str]]: a list containing tuples for each
        video component's university_video_id and key

    Returns:
        Dict[str, List[VideoInfo]]: mapping from university_video_id
        to a list of all VideoInfo objects that have equal values at university_video_id

    """
    print("loading and validating videos")
    video_info_dict = defaultdict(list)
    # check component mp4 exist in S3
    def thread_helper(x):
        video_info = get_video_info(s3, bucket_name, x, error_message)
        if video_info != None:
            lock.acquire()
            video_info_dict[x[0]].append(video_info)
            lock.release()

    def run(thread_helper, video_info_param):
        with ThreadPoolExecutor(max_workers=15) as pool:
            results = list(
                tqdm(
                    pool.map(thread_helper, video_info_param),
                    total=len(video_info_param),
                )
            )
        return results

    run(thread_helper, video_info_param)
    return video_info_dict


def _validate_mp4(
    video_info_dict: Dict[str, List[VideoInfo]], error_message: List[ErrorMessage]
) -> None:
    """
    This function checks a set of MP4 files on S3 for the following properties:
    - null fps
    - consistent width/height
    - rotation consistent
    - video time base consistent
    - video codec consistent
    - audio codec consistent
    - mp4 duration too large or small
    - mp4 duration does not exist
    - video fps consistent
    - SAR consistent
    - FPS consistent

    Args:
        video_info_dict: Dict[str, List[VideoInfo]]: mapping from university_video_id
        to a list of all VideoInfo objects that have equal values at university_video_id

        error_message: List[ErrorMessage]: a list to store ErrorMessage objects
        generated when validating mp4 files in s3.
    """
    print("validating mp4")
    for video_id, video_infos in tqdm(video_info_dict.items()):
        total_vcodec = set()
        total_acodec = set()
        total_rotate = set()
        total_dimension = set()
        total_size = set()
        total_vtb = set()
        total_sar = set()
        total_fps = set()

        for i, video_info in enumerate(video_infos):
            # check consistent video codec
            if video_info.vcodec != None:
                total_vcodec.add(video_info.vcodec)

            # check consistent audio time base
            if video_info.acodec != None:
                total_acodec.add(video_info.acodec)

            # check consistent rotation
            if video_info.rotate != None:
                total_rotate.add(video_info.rotate)
            else:
                error_message.append(
                    ErrorMessage(
                        video_id,
                        "rotation_info_not_found_warning",
                        f"component {i} has no rotation information",
                    )
                )

            # check consistent width and height
            if video_info.sample_width != None and video_info.sample_height != None:
                total_size.add((video_info.sample_width, video_info.sample_height))

                if (
                    video_info.sample_width < video_info.sample_height
                    and video_info.rotate == None
                ):
                    error_message.append(
                        ErrorMessage(
                            video_id,
                            "component_having_width_lt_height_error",
                            f"component {i} has width < height without rotation",
                        )
                    )

            # check consistent video time base
            if video_info.video_time_base != None:
                total_vtb.add(video_info.video_time_base)

            # check consistent sar
            if video_info.sar != None:
                total_sar.add(video_info.sar)

            # check null/inconsistent video fps
            if video_info.fps == None:
                error_message.append(
                    ErrorMessage(
                        video_id,
                        "missing_fps_info_warning",
                        f"component {i} has null fps value",
                    )
                )
            else:
                total_fps.add(video_info.fps)

            # check null mp4 duration
            if video_info.mp4_duration == None:
                error_message.append(
                    ErrorMessage(
                        video_id,
                        "missing_mp4_duration_info_warning",
                        f"component {i} has no mp4 duration",
                    )
                )
            else:
                video_length = (
                    video_info.vstart + video_info.vduration
                    if video_info.vstart != None
                    else video_info.vduration
                )
                audio_length = (
                    video_info.astart + video_info.aduration
                    if video_info.astart != None
                    else video_info.aduration
                )
                delta = abs(max(video_length, audio_length) - video_info.mp4_duration)
                if delta > 0.5:
                    error_message.append(
                        ErrorMessage(
                            video_id,
                            "mp4_duration_too_large_or_small_warning",
                            f"component {i} has an mp4 duration that's too large or small",
                        )
                    )

        if len(total_vcodec) > 1:
            error_message.append(
                ErrorMessage(
                    video_id,
                    "inconsistent_video_codec_error",
                    "inconsistent video codec",
                )
            )
        if len(total_acodec) > 1:
            error_message.append(
                ErrorMessage(
                    video_id,
                    "inconsistent_audio_codec_error",
                    "inconsistent audio codec",
                )
            )
        if len(total_rotate) > 1:
            error_message.append(
                ErrorMessage(
                    video_id, "inconsistent_rotation_error", "inconsistent rotation"
                )
            )
        if len(total_size) > 1:
            error_message.append(
                ErrorMessage(
                    video_id,
                    "inconsistent_width_height_pair_error",
                    "components with inconsistent width x height",
                )
            )
        if len(total_vtb) > 1:
            error_message.append(
                ErrorMessage(
                    video_id,
                    "inconsistent_video_time_base_error",
                    "inconsistent video time base",
                )
            )
        if len(total_sar) > 1:
            error_message.append(
                ErrorMessage(video_id, "inconsistent_sar_warning", "inconsistent sar")
            )
        if len(total_fps) > 1:
            error_message.append(
                ErrorMessage(
                    video_id, "inconsistent_video_fps_warning", "inconsistent video fps"
                )
            )


def _validate_synchronized_videos(
    video_metadata_dict: Dict[str, VideoMetadata],
    synchronized_video_dict: Dict[str, SynchronizedVideos],
    error_message: List[ErrorMessage],
) -> None:
    """
    Args:
        synchronized_video_dict: Dict[str, SynchronizedVideos]: mapping from
        video_grouping_id  to a list of all SynchronizedVideos objects that
        have equal values at video_grouping_id

        error_message: List[ErrorMessage]: a list to store ErrorMessage objects
        generated when validating synchronized_videos.csv.
    """
    print("validating syncrhonized videos")
    if synchronized_video_dict:
        for video_grouping_id, components in synchronized_video_dict.items():
            for component in components:
                for video_id, num in component.associated_videos:
                    if video_id not in video_metadata_dict:
                        error_message.append(
                            ErrorMessage(
                                video_id,
                                "video_not_found_in_video_metadata_error",
                                f"{video_id} in synchronized_video_dict can't be found in video_metadata",
                            )
                        )


def _validate_auxilliary_videos(
    video_metadata_dict: Dict[str, VideoMetadata],
    auxiliary_video_component_dict: Dict[str, List[AuxiliaryVideoComponentDataFile]],
    component_types: Dict[str, ComponentType],
    error_message: List[ErrorMessage],
) -> None:
    """
    Args:
        auxiliary_video_component_dict: Dict[str, List[AuxiliaryVideoComponentDataFile]]:
        mapping from university_video_id to a list of all AuxiliaryVideoComponentDataFile
        objects that have equal values at university_video_id

        error_message: List[ErrorMessage]: a list to store ErrorMessage objects
        generated when validating auxilliary_videos.csv.
    """
    # Check ids in auxiliary_video_component_dict are in video_metadata_dict
    # and that the component_type is valid
    print("validating auxiliary videos")
    if auxiliary_video_component_dict:
        for video_id, components in auxiliary_video_component_dict.items():
            if video_id not in video_metadata_dict:
                error_message.append(
                    ErrorMessage(
                        video_id,
                        "video_not_found_in_video_metadata_error",
                        f"{video_id} in auxiliary_video_component_dict can't be found in video_metadata",
                    )
                )
            elif video_metadata_dict[video_id].number_video_components != len(
                components
            ):
                error_message.append(
                    ErrorMessage(
                        video_id,
                        "video_component_length_inconsistent_error",
                        f"the video has {len(components)} auxiliary components when it"
                        f" should have {video_metadata_dict[video_id].number_video_components}",
                    )
                )
            components.sort(key=lambda x: x.component_index)
            for i in range(len(components)):
                component = components[i]
                if i != component.component_index:
                    error_message.append(
                        ErrorMessage(
                            video_id,
                            "video_component_wrong_index_error",
                            f"the video component has auxiliary component index {component.component_index}"
                            f" when it should have {i}",
                        )
                    )
                if component.component_type_id not in component_types:
                    error_message.append(
                        ErrorMessage(
                            video_id,
                            "component_type_id_not_found_error",
                            f"auxiliary component's component_type_id: '{component.component_type_id} does not exist in component_types'",
                        )
                    )


def _validate_participant(
    video_metadata_dict: Dict[str, VideoMetadata],
    participant_dict: Dict[str, Particpant],
    error_message: List[ErrorMessage],
) -> None:
    """
    Args:
        participant_dict: Dict[str, Particpant]: mapping from participant_id
        to Participant objects that have equal values at participant_id

        error_message: List[ErrorMessage]: a list to store ErrorMessage objects
        generated when validating participants.csv.
    """
    print("validating participants")
    if participant_dict:
        for participant_id, components in participant_dict.items():
            if participant_id not in video_metadata_dict:
                error_message.append(
                    ErrorMessage(
                        video_id,
                        "participant_not_found_error",
                        f" participant '{participant_id}' not exist in "
                        " video metadata",
                    )
                )


def _validate_annotations(
    video_metadata_dict: Dict[str, VideoMetadata],
    annotations_dict: Dict[str, Annotations],
    error_message: List[ErrorMessage],
) -> None:
    """
    Args:
        annotations_dict: Dict[str, Annotations]: mapping from participant_id
        to Participant objects that have equal values at participant_id

        error_message: List[ErrorMessage]: a list to store ErrorMessage objects
        generated when validating participants.csv.
    """
    print("validating annotations")
    if annotations_dict:
        for video_id in annotations_dict:
            if video_id not in video_metadata_dict:
                error_message.append(
                    ErrorMessage(
                        video_id,
                        "video_not_found_in_video_metadata_error",
                        f"{video_id} in annotations_dict can't be found in video_metadata",
                    )
                )


def _validate_video_metadata(
    video_metadata_dict: Dict[str, VideoMetadata],
    video_components_dict: Dict[str, List[VideoComponentFile]],
    participant_dict: Dict[str, Particpant],
    physical_setting_dict: Dict[str, PhysicalSetting],
    devices: Dict[str, Device],
    component_types: Dict[str, ComponentType],
    scenarios: Dict[str, Scenario],
    error_message: List[ErrorMessage],
) -> None:
    """
    Args:
        video_metadata_dict: Dict[str, VideoMetadata]: mapping from
        university_video_id to a list of all VideoMetaData objects that
        have equal values at university_video_id

        video_components_dict: Dict[str, List[VideoComponentFile]]:
        mapping from university_video_id to a list of all VideoComponentFile
        objects that have equal values at university_video_id

        participant_dict: Dict[str, Particpant]: mapping from participant_id
        to Participant objects that have equal values at participant_id

        physical_setting_dict: Dict[str, PhysicalSetting],
        devices: Dict[str, Device],
        component_types: Dict[str, ComponentType],
        scenarios: Dict[str, Scenario],

        error_message: List[ErrorMessage]: a list to store ErrorMessage objects
        generated when validating video_metadata.csv.
    """
    print("validating video metadata")
    # Check from video_metadata:
    video_ids = set()
    for video_id, video_metadata in video_metadata_dict.items():
        #   0. check no duplicate university_video_id
        if video_id in video_ids:
            error_message.append(
                ErrorMessage(
                    video_id,
                    "duplicate_video_id_error",
                    f"duplicate video id in metadata",
                )
            )
        video_ids.add(video_id)
        #   1. participant_ids are in participant_dict
        if participant_dict:
            if video_metadata.recording_participant_id not in participant_dict:
                error_message.append(
                    ErrorMessage(
                        video_id,
                        "participant_id_not_found_error",
                        f"recording_participant_id '{video_metadata.recording_participant_id}' not in participant_dict",
                    )
                )
            if video_metadata.recording_participant_id is None:
                error_message.append(
                    ErrorMessage(
                        video_id, "null_participant_id_warning", "null participant_id"
                    )
                )

        #   2. scenario_ids are in scenarios
        for scenario_id in video_metadata.video_scenario_ids:
            if scenario_id not in scenarios:
                error_message.append(
                    ErrorMessage(
                        video_id,
                        "scenario_id_not_found_error",
                        f"video_scenario_id: '{scenario_id}' not in scenarios.csv",
                    )
                )

        #   3. device_ids are in devices
        if video_metadata.device_id and video_metadata.device_id not in devices:
            error_message.append(
                ErrorMessage(
                    video_id,
                    "device_id_not_found_error",
                    f"device_id '{video_metadata.device_id}' not in devices.csv",
                )
            )

        #   4. physical_settings are in physical_setting
        if physical_setting_dict:
            if (
                video_metadata.physical_setting_id
                and video_metadata.physical_setting_id not in physical_setting_dict
            ):
                error_message.append(
                    ErrorMessage(
                        video_id,
                        "physical_setting_id_not_found_error",
                        f"physical_setting_id '{video_metadata.physical_setting_id}' not in physical_setting.csv",
                    )
                )

        #   5. university_video_ids are in components
        if video_id not in video_components_dict:
            error_message.append(
                ErrorMessage(
                    video_id,
                    "video_not_found_in_video_components_error",
                    f"{video_id} in video_metadata can't be found in video_components_dict",
                )
            )


def validate_university_files(  # noqa :C901
    video_metadata_dict: Dict[str, VideoMetadata],
    video_components_dict: Dict[str, List[VideoComponentFile]],
    auxiliary_video_component_dict: Dict[str, List[AuxiliaryVideoComponentDataFile]],
    participant_dict: Dict[str, Particpant],
    synchronized_video_dict: Dict[str, SynchronizedVideos],
    physical_setting_dict: Dict[str, PhysicalSetting],
    annotations_dict: Dict[str, Annotations],
    devices: Dict[str, Device],
    component_types: Dict[str, ComponentType],
    scenarios: Dict[str, Scenario],
    bucket_name: str,
    s3: botocore.client.BaseClient,
    error_details_path: str,
    error_summary_path: str,
) -> List[ErrorMessage]:
    error_message = []
    # Check ids in video_components_dict are in video_metadata_dict
    # and the # of components is correct
    s3_bucket = S3Helper(s3, bucket_name)
    video_info_param = []

    video_info_param = _validate_video_components(
        s3, bucket_name, video_metadata_dict, video_components_dict, error_message
    )
    video_info_dict = _get_videos(s3, bucket_name, video_info_param, error_message)
    _validate_mp4(video_info_dict, error_message)
    _validate_synchronized_videos(
        video_metadata_dict, synchronized_video_dict, error_message
    )
    _validate_auxilliary_videos(
        video_metadata_dict,
        auxiliary_video_component_dict,
        component_types,
        error_message,
    )
    _validate_participant(video_metadata_dict, participant_dict, error_message)
    _validate_annotations(video_metadata_dict, annotations_dict, error_message)
    _validate_video_metadata(
        video_metadata_dict,
        component_types,
        participant_dict,
        scenarios,
        devices,
        physical_setting_dict,
        video_components_dict,
        error_message,
    )

    error_dict = collections.defaultdict(int)
    if error_message:
        for err in error_message:
            error_dict[err.errorType] += 1

    fields = ["univeristy_video_id", "errorType", "description"]
    with open(error_details_path, "w") as f:
        # using csv.writer method from CSV package
        write = csv.writer(f)
        write.writerow(fields)
        for e in error_message:
            write.writerow([e.uid, e.errorType, e.description])
    with open(error_summary_path, "w") as f:
        write = csv.writer(f)
        write.writerow(["error_type", "num_of_occurrences"])
        for error_type, error_counts in error_dict.items():
            write.writerow([error_type, error_counts])
    return error_message


def validate_all(
    path, s3, standard_metadata_folder, error_details_path, error_summary_path
):
    # get access to metadata_folder
    devices, component_types, scenarios = load_standard_metadata_files(
        s3, path, standard_metadata_folder
    )
    bucket, path = split_s3_path(path)

    (
        video_metadata_dict,
        video_components_dict,
        auxiliary_video_component_dict,
        participant_dict,
        synchronized_video_dict,
        physical_setting_dict,
        annotations_dict,
    ) = load_university_files(s3, bucket, path)

    validate_university_files(
        video_metadata_dict,
        video_components_dict,
        auxiliary_video_component_dict,
        participant_dict,
        synchronized_video_dict,
        physical_setting_dict,
        annotations_dict,
        devices,
        component_types,
        scenarios,
        bucket,
        s3,
        error_details_path,
        error_summary_path,
    )



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

"""
A data object for storing user options, along with utilities for parsing input options
from command line flags and a configuration file.
"""
import argparse
import json
from pathlib import Path
from typing import List, NamedTuple, Set

import boto3.session
from botocore.exceptions import ProfileNotFound
from ego4d.cli.universities import UNIV_TO_BUCKET

unis = [
    "unict",
    "cmu",
    "iiith",
    "minnesota",
    "utokyo",
    "kaust",
    "indiana",
    "nus",
    "bristol",
    "georgiatech",
    "frl_track_1_public",
    "cmu_africa",
    "uniandes",
]
meta_path = [
    "metadata_v5",
    "metadata_v27",
    "2021-08-29_august_fixed",
    "metadata_v10",
    "metadata_v3",
    "metadata_v7",
    "metadata-v4",
    "metadata_v1.2",
    "metadata-v3",
    "metadata_v3",
    "track1/metadata_v0",
    "metadata_v3",
    "metadata_v0",
]


class ValidatedConfig(NamedTuple):
    """
    Data object that stores validated user-supplied configuration options for a video
    download operation.
    """

    input_directory: str
    validate_all: bool
    metadata_folder: str
    error_details_name: str
    error_summary_name: str
    aws_profile_name: str
    universities: Set[str]


class Config(NamedTuple):
    """
    Data object that stores the user-supplied configuration options for a video download
    operation.
    """

    input_directory: str
    validate_all: bool
    metadata_folder: str
    error_details_name: str
    error_summary_name: str
    aws_profile_name: str = "default"
    universities: List[str] = []


def validate_config(cfg: Config) -> ValidatedConfig:
    """

    Args:
        cfg: A user-supplied configuration for the download operation
    Returns:
        A ValidatedConfig if all user-supplied options appear valid
    """

    try:
        boto3.session.Session(profile_name=cfg.aws_profile_name)
    except ProfileNotFound:
        raise RuntimeError(f"Could not find AWS profile '{cfg.aws_profile_name}'.")

    return ValidatedConfig(
        input_directory=cfg.input_directory,
        validate_all=bool(cfg.validate_all),
        metadata_folder=cfg.metadata_folder,
        error_details_name=cfg.error_details_name,
        error_summary_name=cfg.error_summary_name,
        aws_profile_name=cfg.aws_profile_name,
        universities=set(cfg.universities) if cfg.universities else {},
    )


def config_from_args(args=None) -> Config:
    """
    Parses command line flags and returns a Config object with corresponding values from
    the flags.
    """
    # Parser for a configuration file
    json_parser = argparse.ArgumentParser(
        description="Command line tool to download Ego4D datasets from Amazon S3"
    )

    json_parser.add_argument(
        "--config_path",
        type=Path,
        help="Local path to a config JSON file. If specified, the flags will be read "
        "from this file instead of the command line.",
    )

    args, remaining = json_parser.parse_known_args(args=args)

    # Parser for command line flags other than the configuration file
    flag_parser = argparse.ArgumentParser(add_help=False)

    # required_flags = {"output_directory"}
    flag_parser.add_argument(
        "-i",
        "--input_directory",
        help="The S3 path where the metadata is stored",
    )

    flag_parser.add_argument(
        "-a",
        "--all",
        default=False,
        help="validate all files in S3",
        dest="validate_all",
    )
    flag_parser.add_argument(
        "-mf",
        "--metadata_folder",
        help="the local metadata folder where csv files (video components, metadata, etc.) will be downloaded to and stored in",
    )
    flag_parser.add_argument(
        "-ed", "--error_details_name", help="output file name for error details"
    )
    flag_parser.add_argument(
        "-es", "--error_summary_name", help="output file name for error summary"
    )
    flag_parser.add_argument(
        "--aws_profile_name",
        help="Defaults to 'default'. Specifies the AWS profile name from "
        "~/.aws/credentials to use for the download",
        default="default",
    )
    flag_parser.add_argument(
        "--universities",
        nargs="+",
        choices=UNIV_TO_BUCKET.keys(),
        help="List of university IDs. If specified, only UIDs from the S3 buckets "
        "belonging to the listed universities will be downloaded. A full list of "
        "university IDs can be found in the ego4d/cli/universities.py file.",
    )

    # Use the values in the config file, but set them as defaults to flag_parser so they
    # can be overridden by command line flags
    if args.config_path:
        with open(args.config_path.expanduser()) as f:
            config_contents = json.load(f)
            flag_parser.set_defaults(**config_contents)

    parsed_args = flag_parser.parse_args(remaining)

    flags = {k: v for k, v in vars(parsed_args).items() if v is not None}

    # Note: Since the flags from the config file are being used as default argparse
    # values, we can't set required=True. Doing so would mean that the user couldn't
    # leave them unspecified when invoking the CLI.

    config = Config(**flags)

    # We need to check the universities values here since they might have been set
    # through the JSON config file, in which case the argparse checks won't validate
    # them
    universities = set(UNIV_TO_BUCKET.keys())
    unrecognized = set(config.universities) - universities
    if unrecognized:
        raise RuntimeError(
            f"Unrecognized universities: {unrecognized}. Please choose "
            f"from: {universities}"
        )

    return config



# pyre-unsafe
import csv
import json
import logging
import os
import tempfile
from collections import defaultdict
from dataclasses import (
    dataclass,
    Field,
    field as dataclass_field,
    fields as dataclass_fields,
)
from datetime import datetime
from threading import Lock
from typing import Any, Callable, Dict, List, Tuple, Union

from ego4d.internal.credential_s3 import S3Helper

from tqdm import tqdm


logging.basicConfig(filename="example.log", level=logging.DEBUG)
lock = Lock()


def split_s3_path(s3_path: str) -> Tuple[str, str]:
    """
    Splits a full s3_path of the form "s3://bucket/folder/.../file.ext"
    into a tuple of ("bucket", "folder/.../file.ext")

    If this doesn't start with s3:// it will assume that the there is no
    bucket and the path is just returned as the second output
    """
    if s3_path[:5] == "s3://":
        s3_path_components = s3_path[5:].split("/")
        return s3_path_components[0], "/".join(s3_path_components[1:])
    # pyre-fixme[7]: Expected `Tuple[str, str]` but got `Tuple[None, str]`.
    return None, s3_path


class DataclassFieldCaster:
    """
    Class to allow subclasses wrapped in @dataclass to automatically
    cast fields to their relevant type by default.

    Also allows for an arbitrary intialization function to be applied
    for a given field.
    """

    COMPLEX_INITIALIZER = "DataclassFieldCaster__complex_initializer"

    def __post_init__(self) -> None:
        f"""
        This function is run by the dataclass library after '__init__'.

        Here we use this to ensure all fields are casted to their declared types
        and to apply any complex field_initializer functions that have been
        declared via the 'complex_initialized_dataclass_field' method of
        this class.

        A complex field_initializer for a given field would be stored in the
        field.metadata dictionary at:
            key = '{self.COMPLEX_INITIALIZER}' (self.COMPLEX_INITIALIZER)

        """
        for field in dataclass_fields(self):
            value = getattr(self, field.name)
            # First check if the datafield has been set to the declared type or
            # if the datafield has a declared complex field_initializer.
            if (
                not isinstance(value, field.type)
                or DataclassFieldCaster.COMPLEX_INITIALIZER in field.metadata
            ):
                # Apply the complex field_initializer function for this field's value,
                # assert that the resultant type is the declared type of the field.
                if DataclassFieldCaster.COMPLEX_INITIALIZER in field.metadata:
                    setattr(
                        self,
                        field.name,
                        field.metadata[DataclassFieldCaster.COMPLEX_INITIALIZER](value),
                    )
                    assert isinstance(getattr(self, field.name), field.type), (
                        f"'field_initializer' function of {field.name} must return "
                        f"type {field.type} but returned type {type(getattr(self, field.name))}"
                    )
                elif field.type == bool:
                    if value:
                        if type(value) == str and value.lower() in ["true", "false"]:
                            value = value.lower() == "true"
                        else:
                            value = bool(int(value))
                    else:
                        value = False
                    setattr(self, field.name, value)
                elif value is not None:
                    # Otherwise attempt to cast the field's value to its declared type.
                    try:
                        value = field.type(value)
                    except ValueError:
                        value = None
                    setattr(self, field.name, value)

    @staticmethod
    def complex_initialized_dataclass_field(
        field_initializer: Callable, **kwargs
    ) -> Field:
        """
        Allows for the setting of a function to be called on the
        named parameter associated with a field during initialization,
        after __init__() completes.

        Args:
            field_initializer (Callable):
                The function to be called on the field

            **kwargs: To be passed downstream to the dataclasses.field method

        Returns:
            (dataclasses.Field) that contains the field_initializer and kwargs infoÎ
        """
        metadata = kwargs.get("metadata") or {}
        assert DataclassFieldCaster.COMPLEX_INITIALIZER not in metadata
        metadata[DataclassFieldCaster.COMPLEX_INITIALIZER] = field_initializer
        kwargs["metadata"] = metadata
        return dataclass_field(**kwargs)


@dataclass
class ErrorMessage:
    """Class for keeping track of the error generated in validation."""

    uid: str
    errorType: str
    description: str

    def _init_(self, uid, error, desc):
        self.uid = uid
        self.errorType = error
        self.description = desc


@dataclass
class Device(DataclassFieldCaster):
    device_id: int
    name: str


@dataclass
class ComponentType(DataclassFieldCaster):
    component_type_id: int
    name: str


@dataclass
class Scenario(DataclassFieldCaster):
    scenario_id: int
    name: str


DATE_FORMAT_STR = "%Y-%m-%d %H:%M:%S"


def DATE_INITIALIZATION_FUNC(a: str):
    # TODO (Support milleseconds if needed)
    # a = a.split(".")[0]
    # Fall back to default if not available
    return (
        datetime.strptime(a.split(".")[0], DATE_FORMAT_STR)
        if a
        else datetime(1900, 1, 1)
    )


def json_load_or_empty_dict(s):
    return json.loads(s) if s else {}


def json_load_or_empty_list(s):
    if s:
        # Some universities double encode thier data as JSON
        # This accounts for that
        while type(s) == str:
            s = json.loads(s)
        return s
    else:
        return []


@dataclass
class VideoMetadata(DataclassFieldCaster):
    university_video_id: str
    university_video_folder_path: str
    number_video_components: int
    # pyre-fixme[8]: Attribute has type `datetime`; used as `Field[typing.Any]`.
    start_date_recorded_utc: datetime = (
        DataclassFieldCaster.complex_initialized_dataclass_field(
            DATE_INITIALIZATION_FUNC
        )
    )
    recording_participant_id: str
    device_id: int
    # pyre-fixme[8]: Attribute has type `Dict[typing.Any, typing.Any]`; used as
    #  `Field[typing.Any]`.
    video_device_settings: dict = (
        DataclassFieldCaster.complex_initialized_dataclass_field(
            json_load_or_empty_dict
        )
    )
    physical_setting_id: str
    # pyre-fixme[8]: Attribute has type `List[typing.Any]`; used as `Field[typing.Any]`.
    video_scenario_ids: list = (
        DataclassFieldCaster.complex_initialized_dataclass_field(  # noqa
            json_load_or_empty_list
        )
    )


@dataclass
class VideoComponentFile(DataclassFieldCaster):
    university_video_id: str
    video_component_relative_path: str
    component_index: int
    is_redacted: bool
    # pyre-fixme[8]: Attribute has type `datetime`; used as `Field[typing.Any]`.
    start_date_recorded_utc: datetime = (
        DataclassFieldCaster.complex_initialized_dataclass_field(
            DATE_INITIALIZATION_FUNC
        )
    )
    # pyre-fixme[8]: Attribute has type `Dict[typing.Any, typing.Any]`; used as
    #  `Field[typing.Any]`.
    compression_settings: dict = (
        DataclassFieldCaster.complex_initialized_dataclass_field(
            json_load_or_empty_dict
        )
    )
    includes_audio: bool
    # pyre-fixme[8]: Attribute has type `Dict[typing.Any, typing.Any]`; used as
    #  `Field[typing.Any]`.
    component_metadata: dict = (
        DataclassFieldCaster.complex_initialized_dataclass_field(  # noqa
            json_load_or_empty_dict
        )
    )
    # pyre-fixme[8]: Attribute has type `Dict[typing.Any, typing.Any]`; used as
    #  `Field[typing.Any]`.
    deidentification_metadata: dict = (
        DataclassFieldCaster.complex_initialized_dataclass_field(
            json_load_or_empty_dict
        )
    )


@dataclass
class AuxiliaryVideoComponentDataFile(DataclassFieldCaster):
    university_video_id: str
    component_index: int
    component_type_id: int
    video_component_relative_path: str


@dataclass
class Particpant(DataclassFieldCaster):
    participant_id: str
    # pyre-fixme[8]: Attribute has type `Dict[typing.Any, typing.Any]`; used as
    #  `Field[typing.Any]`.
    participant_metadata: dict = (
        DataclassFieldCaster.complex_initialized_dataclass_field(
            json_load_or_empty_dict
        )
    )


@dataclass
class SynchronizedVideos(DataclassFieldCaster):
    video_grouping_id: str
    # pyre-fixme[8]: Attribute has type `Dict[typing.Any, typing.Any]`; used as
    #  `Field[typing.Any]`.
    synchronization_metadata: dict = (
        DataclassFieldCaster.complex_initialized_dataclass_field(
            json_load_or_empty_dict
        )
    )
    # pyre-fixme[8]: Attribute has type `Dict[typing.Any, typing.Any]`; used as
    #  `Field[typing.Any]`.
    associated_videos: dict = (
        DataclassFieldCaster.complex_initialized_dataclass_field(  # noqa
            json_load_or_empty_dict
        )
    )


@dataclass
class PhysicalSetting(DataclassFieldCaster):
    setting_id: str
    name: str
    associated_matterport_scan_path: str


@dataclass
class Annotations(DataclassFieldCaster):
    university_video_id: str
    start_seconds: float
    end_seconds: float
    # pyre-fixme[8]: Attribute has type `Dict[typing.Any, typing.Any]`; used as
    #  `Field[typing.Any]`.
    annotation_data: dict = (
        DataclassFieldCaster.complex_initialized_dataclass_field(  # noqa
            json_load_or_empty_dict
        )
    )


def load_dataclass_dict_from_csv(
    input_csv_file_path: str,
    dataclass_class: type,
    dict_key_field: str,
    list_per_key: bool = False,
) -> Dict[Any, Union[Any, List[Any]]]:
    """
    Args:
        input_csv_file_path (str): File path of the csv to read from
        dataclass_class (type): The dataclass to read each row into.
        dict_key_field (str): The field of 'dataclass_class' to use as
            the dictionary key.
        list_per_key (bool) = False: If the output data structure
        contains a list of dataclass objects per key, rather than a
        single unique dataclass object.

    Returns:
        Dict[Any, Union[Any, List[Any]] mapping from the dataclass
        value at attr = dict_key_field to either:

        if 'list_per_key', a list of all dataclass objects that
        have equal values at attr = dict_key_field, equal to the key

        if not 'list_per_key', the unique dataclass object
        for which the value at attr = dict_key_field is equal to the key

    Raises:
        AssertionError: if not 'list_per_key' and there are
        dataclass obejcts with equal values at attr = dict_key_field
    """

    output_dict = defaultdict(list) if list_per_key else {}
    with open(input_csv_file_path, newline="") as csvfile:
        reader = csv.reader(csvfile, delimiter=",", quotechar='"')
        column_index = {header: i for i, header in enumerate(next(reader))}
        for line in tqdm(reader):
            datum = dataclass_class(
                *(
                    line[column_index[field.name]]
                    for field in dataclass_fields(dataclass_class)
                    if field.name in column_index
                )
            )
            dict_key = getattr(datum, dict_key_field)
            if list_per_key:
                output_dict[dict_key].append(datum)
            else:
                assert (
                    dict_key not in output_dict
                ), f"Multiple entries for {dict_key} in {csvfile}"
                output_dict[dict_key] = datum
    return output_dict


def load_standard_metadata_files(
    s3,
    path: str,
    standard_metadata_folder: str,
) -> Tuple[Dict[str, Device], Dict[str, ComponentType], Dict[str, Scenario]]:
    # Load consortium-wide metadata into memory
    bucket, path = split_s3_path(path)
    # uncomment the below line to download metadata to the specified metadata folder
    # download_standard_metadata_files_from_s3(s3, bucket, path, standard_metadata_folder)
    devices = load_dataclass_dict_from_csv(
        f"{standard_metadata_folder}/device.csv", Device, "device_id"
    )
    component_types = load_dataclass_dict_from_csv(
        f"{standard_metadata_folder}/component_type.csv",
        ComponentType,
        "component_type_id",
    )
    scenarios = load_dataclass_dict_from_csv(
        f"{standard_metadata_folder}/scenario.csv", Scenario, "scenario_id"
    )
    validate_standard_metadata_files(devices, component_types, scenarios)
    return devices, component_types, scenarios


def download_standard_metadata_files_from_s3(
    s3, bucket: str, path: str, standard_metadata_folder: str
) -> Tuple[Dict[str, Device], Dict[str, ComponentType], Dict[str, Scenario]]:
    s3_bucket = S3Helper(s3, bucket)
    auxiliary_videos = s3_bucket.get_file(
        os.path.join(path, "auxiliary_video_component_data_file.csv"),
        os.path.join(
            standard_metadata_folder + "auxiliary_video_component_data_file.csv"
        ),
    )
    physical_settings = s3_bucket.get_file(
        os.path.join(path, "physical_setting.csv"),
        os.path.join(standard_metadata_folder + "physical_setting.csv"),
    )
    video_components = s3_bucket.get_file(
        os.path.join(path, "video_component_file.csv"),
        os.path.join(standard_metadata_folder + "video_component_file.csv"),
    )
    video_metadata = s3_bucket.get_file(
        os.path.join(path, "video_metadata.csv"),
        os.path.join(standard_metadata_folder + "video_metadata.csv"),
    )
    return auxiliary_videos, physical_settings, video_components, video_metadata


def validate_standard_metadata_files(
    devices: Dict[str, Device],
    component_types: Dict[str, ComponentType],
    scenarios: Dict[str, Scenario],
) -> None:
    # At the time of writing code these are the lengths of the files
    assert len(devices) >= 8
    assert len(component_types) >= 4
    assert len(scenarios) >= 108


def load_university_files(
    s3,
    s3_bucket_name: str,
    metadata_folder_prefix: str,
) -> Tuple[
    Dict[str, VideoMetadata],
    Dict[str, List[VideoComponentFile]],
    Dict[str, List[AuxiliaryVideoComponentDataFile]],
    Dict[str, Particpant],
    Dict[str, SynchronizedVideos],
    Dict[str, PhysicalSetting],
    Dict[str, Annotations],
]:

    if s3_bucket_name:
        s3_bucket = S3Helper(s3, s3_bucket_name)
        truncated, available_metadata_files = s3_bucket.ls(metadata_folder_prefix)
        assert not truncated, (
            f"Folder {s3_bucket_name}/{metadata_folder_prefix} " "has too many entries"
        )
        available_files = {f.key.split("/")[-1] for f in available_metadata_files}
    else:
        available_files = os.listdir(metadata_folder_prefix)

    with tempfile.TemporaryDirectory("consortium_ingester") as tempdir:
        # Load reqiured files
        # Load video_metadata.csv
        file_name = "video_metadata.csv"
        assert file_name in available_files, (
            f"required file {file_name} not found in "
            f"{s3_bucket_name}/{metadata_folder_prefix}"
        )
        if s3_bucket_name:
            local_file_path = f"{tempdir}/{file_name}"
            s3_bucket.get_file(f"{metadata_folder_prefix}/{file_name}", local_file_path)
        else:
            local_file_path = f"{metadata_folder_prefix}/{file_name}"
        video_metadata_dict = load_dataclass_dict_from_csv(
            local_file_path,
            VideoMetadata,
            "university_video_id",
        )

        # Load video_component_file.csv
        file_name = "video_component_file.csv"
        assert file_name in available_files, (
            f"required file {file_name} not found in "
            f"{s3_bucket_name}/{metadata_folder_prefix}"
        )
        if s3_bucket_name:
            local_file_path = f"{tempdir}/{file_name}"
            s3_bucket.get_file(f"{metadata_folder_prefix}/{file_name}", local_file_path)
        else:
            local_file_path = f"{metadata_folder_prefix}/{file_name}"
        video_components_dict = load_dataclass_dict_from_csv(
            local_file_path,
            VideoComponentFile,
            "university_video_id",
            list_per_key=True,
        )

        # Load optional files
        # Load auxiliary_video_component_data_file.csv, if available
        file_name = "auxiliary_video_component_data_file.csv"
        auxiliary_video_component_dict = {}
        if file_name in available_files:
            if s3_bucket_name:
                local_file_path = f"{tempdir}/{file_name}"
                s3_bucket.get_file(
                    f"{metadata_folder_prefix}/{file_name}", local_file_path
                )
            else:
                local_file_path = f"{metadata_folder_prefix}/{file_name}"
            auxiliary_video_component_dict = load_dataclass_dict_from_csv(
                local_file_path,
                AuxiliaryVideoComponentDataFile,
                "university_video_id",
                list_per_key=True,
            )

        # Load participant.csv, if available
        file_name = "participant.csv"
        participant_dict = {}
        if file_name in available_files:
            if s3_bucket_name:
                local_file_path = f"{tempdir}/{file_name}"
                s3_bucket.get_file(
                    f"{metadata_folder_prefix}/{file_name}", local_file_path
                )
            else:
                local_file_path = f"{metadata_folder_prefix}/{file_name}"
            participant_dict = load_dataclass_dict_from_csv(
                local_file_path,
                Particpant,
                "participant_id",
            )

        # Load synchronized_videos.csv, if available
        file_name = "synchronized_videos.csv"
        synchronized_video_dict = {}
        if file_name in available_files:
            if s3_bucket_name:
                local_file_path = f"{tempdir}/{file_name}"
                s3_bucket.get_file(
                    f"{metadata_folder_prefix}/{file_name}", local_file_path
                )
            else:
                local_file_path = f"{metadata_folder_prefix}/{file_name}"
            synchronized_video_dict = load_dataclass_dict_from_csv(
                local_file_path,
                SynchronizedVideos,
                "video_grouping_id",
            )

            # Load physical_setting.csv, if available
        file_name = "physical_setting.csv"
        physical_setting_dict = {}
        if file_name in available_files:
            if s3_bucket_name:
                local_file_path = f"{tempdir}/{file_name}"
                s3_bucket.get_file(
                    f"{metadata_folder_prefix}/{file_name}", local_file_path
                )
            else:
                local_file_path = f"{metadata_folder_prefix}/{file_name}"
            physical_setting_dict = load_dataclass_dict_from_csv(
                local_file_path,
                PhysicalSetting,
                "setting_id",
            )

            # Load annotations.csv, if available
        file_name = "annotations.csv"
        annotations_dict = {}
        if file_name in available_files:
            if s3_bucket_name:
                local_file_path = f"{tempdir}/{file_name}"
                s3_bucket.get_file(
                    f"{metadata_folder_prefix}/{file_name}", local_file_path
                )
            else:
                local_file_path = f"{metadata_folder_prefix}/{file_name}"
            annotations_dict = load_dataclass_dict_from_csv(
                local_file_path,
                Annotations,
                "university_video_id",
            )

    return (
        video_metadata_dict,
        video_components_dict,
        auxiliary_video_component_dict,
        participant_dict,
        synchronized_video_dict,
        physical_setting_dict,
        annotations_dict,
    )



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

"""
Command line tool to download Ego4D datasets.

Examples:
      python -m ego4d.internal.cli \
        -i "s3://ego4d-unict/metadata_v7" \
        -mf "./ego4d/internal/standard_metadata_v10" \
        -ed "error_details" \
        -es "error_summary" \
"""
import boto3
from ego4d.cli.universities import UNIV_TO_BUCKET
from ego4d.internal.config import (
    Config,
    config_from_args,
    meta_path,
    unis,
    validate_config,
)
from ego4d.internal.validate import validate_all


def main_cfg(cfg: Config) -> None:

    validated_cfg = validate_config(cfg)

    # This service resource in the default session will be used for general light-weight
    # requests on the main thread, such as downloading the video manifests and getting
    # S3 object metadata
    # s3 = boto3.session.Session(profile_name=validated_cfg.aws_profile_name).resource(
    #     "s3"
    # )
    s3 = boto3.client("s3")
    if cfg.validate_all:
        for u in unis:
            path = f"s3://{UNIV_TO_BUCKET[u]}/{meta_path[u]}"
            validate_all(
                path,
                s3,
                validated_cfg.metadata_folder,
                validated_cfg.error_details_name,
                validated_cfg.error_summary_name,
            )
    else:
        validate_all(
            validated_cfg.input_directory,
            s3,
            validated_cfg.metadata_folder,
            validated_cfg.error_details_name,
            validated_cfg.error_summary_name,
        )


def main() -> None:
    config = config_from_args()
    main_cfg(config)


if __name__ == "__main__":
    main()
